/*
 * This file is part of XViewdata - a viewdata terminal emulator.
 * Copyright (C) 1997 Peter Corlett. All Rights Reserved.
 *
 * CharMap.h: Header file for character map functions
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef CHARMAP_H
#define CHARMAP_H

#include <iostream.h>

#include "CharCell.h"

/**

  A @var{CharMap} is a representation of a screen display. In addition to
  being a variable size grid of @var{CharCells}, it also knows the current
  cursor position, whether it is visible, and the current `blink' state of
  both the cursor and the flashing characters.

 */

class CharMap {
  /// Width of character map.
  int width;
  /// Height of character map.
  int height;
  /// Cursor X position.
  int cursx;
  /// Cursor Y position.
  int cursy;
  /// TRUE if cursor is visible.
  BOOL cursv;
  /// TRUE if flashing text in `hide' state.
  BOOL flashcycle;
  /// TRUE if flashing cursor in `hide' state.
  BOOL curscycle;
  /// Pointer to character map itself.
  CharCell *cmap;

  ///
  void clear(void);
  ///
  void free(void);

  /**
    Allocates a character map with the given size.
    @param x width of character map.
    @param y height of character map.
    @return FALSE on failure, in which case the @var{CharMap} is invalid.
    */
  BOOL alloc(int x, int y);
  ///
public:
  /// Default constructor.
  CharMap(void);
  /// Copy constructor.
  CharMap(const CharMap &);
  /// Creates a new @var{x} by @var{y} @var{CharMap}.
  CharMap(int x, int y) { alloc(x, y); };
  /// Destructor.
  ~CharMap();
  
  /// Assignment operator.
  CharMap &operator=(const CharMap &);
  
  /// @return nonzero if the two @var{CharMap}s are identical.
  friend BOOL operator==(const CharMap &, const CharMap &);

  /// @return nonzero if the two @var{CharMap}s differ.
  friend BOOL operator!=(const CharMap &, const CharMap &);

  /// @return nonzero if object is valid.
  operator void *(void) const {
    return cmap;
  }

  /// @return nonzero if object is invalid.
  BOOL operator!(void) const {
    return (*this)?FALSE:TRUE;
  };

  /** Resizes the character map. The current contents of the map will be lost.
    @param x new width of character map.
    @param y new height of character map.
    @return FALSE on failure, in which case the @var{CharMap} is invalid.
    */
  BOOL setSize(int x, int y) {
    if((x!=width)||(y!=height)) {
      free(); clear();
      return alloc(x, y);
    }
    return TRUE;
  };
  /// @return the width of the character map.
  int getWidth(void) const {return width;};
  /// @return the height of the character map.
  int getHeight(void) const {return height;};
  /// @return the @var{n}th @var{CharCell} in the character map.
  CharCell &getCharCell(int n=0) const {return cmap[n];};
  /// @return the @var{CharCell} at position (@var{x},@var{y}) in the character map.
  CharCell &getCharCell(int x, int y) const {return cmap[y*getWidth()+x];};
  /// Sets the @var{n}th CharCell to @var{c}.
  void setCharCell(int n, const CharCell &c) {cmap[n]=c;};
  /// Sets the @var{CharCell} at position (@var{x},@var{y}) to @var{c}.
  void setCharCell(int x, int y, const CharCell &c) {cmap[y*getWidth()+x]=c;};

  /// @return the cursor X position.
  int getCursorX(void) const {return cursx;};
  /// @return the cursor Y position.
  int getCursorY(void) const {return cursy;};
  /// @return TRUE if the cursor is visible.
  BOOL getCursorVisibility(void) const {return cursv;};
  /// Sets the cursor X position to @var{x}.
  void setCursorX(int x) {cursx=x;};
  /// Sets the cursor Y position to @var{y}.
  void setCursorY(int y) {cursy=y;};
  /// Sets the cursor visibility flag.
  void setCursorVisibility(BOOL n) {cursv=n;};

  /// Sets the cursor blink cycle.
  void setCursorCycle(BOOL n) {curscycle=n;};
  /// Sets the flash blink cycle.
  void setFlashCycle(BOOL n) {flashcycle=n;};
  /// @return the cursor blink cycle.
  BOOL getCursorCycle(void) const {return curscycle;};
  /// @return the flash blink cycle.
  BOOL getFlashCycle(void) const {return flashcycle;};

  /// Show contents of the object.
  friend ostream &operator<<(ostream &s, const CharCell &c) {
    s << "(CharMap: " << width << ", " << height << ", " << cursx << ", " << cursy << ", "
      << flashcycle << ", " << curscycle << ", ";
    for(int n=0;n<width*height;n++)
      s << charmap[n];
    return s;
  };

  /**

    This method compares the @var{line}th line of this @var{CharMap}
    against @var{c2}. Rather than a true bitwise comparison, it decides
    whether the two lines would look the same when rendered. So if text
    happens to be hidden (because it is in a flash cycle) and the text was
    changed (but still invisible), then the comparison would succeed, since
    the visual effect would be the same for both. On the other hand, just
    having the cursor blink on the line would cause the comparison to fail.

    @return nonzero if the @var{line}th line of @var{CharMap} @var{c2}
    visually matches this one.

*/
  BOOL CompareLines(const CharMap &c2, int line) const;
};

#endif /* CHARMAP_H */

