/*
 * This file is part of XViewdata - a viewdata terminal emulator.
 * Copyright (C) 1997 Peter Corlett. All Rights Reserved.
 *
 * main.cc: Main entry for program.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/time.h>

#include <iostream.h>
#include <String.h>

// libg++
#include <GetOpt.h>
// undef putc from GetOpt.h
#undef putc

#include "../config.h"
#include "global.h"
#include "Viewdata.h"
#include "DisplayDriver.h"
#include "ConnectionDriver.h"
#include "Settings.h"

Settings Settings;

void show_options(void) {
  cerr
    << "Usage: " PROGNAME " [options] server\n"
    << "\n"
    << "Options:\n"
    << " -d s\tSet display options, use default display (default \"" << Settings.displayarg << "\")\n"
    << " -D S:s\tSet display and options (default \"" << Settings.displaydriver << ":" << Settings.displayarg <<"\")\n"
    << " -t s\tSet terminal options, use default terminal (default \"" << Settings.terminalarg << "\")\n"
    << " -T S:s\tSet terminal and options (default \"" << Settings.terminaldriver << ":" << Settings.terminalarg << "\")\n"
    << " -S s\tSet server type (default \"" <<Settings.connectiondriver << "\")\n"
    << " server\tUse this server\n"
    << "\n"
    << "Infrequently used options:\n"
    << " -f s\tSet font path (default \"" << Settings.fontdir << "\")\n"
    << " -N <n>\tSet network buffer size to n (default " << Settings.netbuf << ")\n"
    << "\n";

  cerr << "Devices available: \n";
  for(int n=0;DisplayDrivers[n].name;n++)
    cerr  << " " << DisplayDrivers[n].name;
  cerr << "\n";

  cerr << "Terminals available: \n";
  for(int n=0;TerminalDrivers[n].name;n++)
    cerr  << " " << TerminalDrivers[n].name;
  cerr << "\n";

  cerr << "Server types available: \n";
  for(int n=0;ConnectionDrivers[n].name;n++)
    cerr  << " " << ConnectionDrivers[n].name;
  cerr << "\n";

}

int main(int argc, char *argv[]) {
  PGFUNC;

  GetOpt getopt(argc, argv, "d:D:t:T:S:f:N:h");
  int option_char;

  cerr << PROGNAME " version " VERSION " (" __DATE__  " " __TIME__ ")\n"
       <<"Copyright (C) 1997 Peter Corlett. All rights reserved.\n"
       << "This software comes with NO WARRANTY: see the file COPYING for details.\n"
       << "\n";
  
  while((option_char=getopt())!=EOF) {
    String arg=getopt.optarg;
    int idx;

    switch(option_char) {
    case 'd':
      Settings.displayarg=arg;
      break;
    case 'D':
      idx=arg.index(":");
      if(idx>=0) {
	Settings.displaydriver=arg.before(idx);
	Settings.displayarg=arg.after(idx);
      } else
	Settings.displaydriver=arg;
      break;

    case 't':
      Settings.terminalarg=arg;
      break;
    case 'T':
      idx=arg.index(":");
      if(idx>=0) {
	Settings.terminaldriver=arg.before(idx);
	Settings.terminalarg=arg.after(idx);
      } else
	Settings.terminaldriver=arg;
      break;

    case 'S':
      Settings.connectiondriver=arg;
      break;

    case 'f':
      Settings.fontdir=arg;
      break;
    case 'N':
      Settings.netbuf=atoi(arg);
      break;

    case 'h':
    default:
      show_options();
      return 1;
    }
  }

int n=getopt.optind;
  if(n<argc) {
    Settings.connectionarg=argv[n];
  } else {
    show_options();
    return 1;
  }

  n++;

  DisplayDriver *display=
    newDisplayDriver(Settings.displaydriver, Settings.displayarg);
  if(!display) {
    cerr << PROGNAME ": I cannot open the display \""
	 << Settings.displaydriver <<":" << Settings.displayarg << "\"\n";
    return 1;
  }

  TerminalDriver *tty=
    newTerminalDriver(Settings.terminaldriver, Settings.terminalarg);
  if(!tty) {
    cerr << PROGNAME ": I can't open the terminal \""
	 << Settings.terminaldriver << ":" << Settings.terminalarg << "\"\n";
    return 1;
  }

  ConnectionDriver *sock=
    newConnectionDriver(Settings.connectiondriver, Settings.connectionarg);
  if(!sock) {
    cerr << PROGNAME ": I cannot connect to the service \""
	 << Settings.connectiondriver <<":"<< Settings.connectionarg << "\"\n";
    return 1;
  }

  display->Main(*sock, *tty);

  return 0;

}
