/*
 * This file is part of XViewdata - a viewdata terminal emulator.
 * Copyright (C) 1997 Peter Corlett. All Rights Reserved.
 *
 * Xlib.cc - X Window terminal driver.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "../config.h"

#ifdef USE_XLIB

#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <errno.h>
#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>

#include <X11/keysym.h>

#include <iostream.h>

#include "global.h"
#include "Xlib.h"
#include "RastFunc.h"
#include "FontEntry.h"
#include "Settings.h"

extern Settings Settings;

void Xlib::clear(void) {
  POFUNC;
  fbuf=NULL;
}

void Xlib::free(void) {
  POFUNC;
  if(fbuf)
    delete fbuf;
  clear();
}

Xlib::Xlib(void) {
  POFUNC;
  clear();
}

Xlib::~Xlib() {
  POFUNC;
  free();
}

BOOL Xlib::EnableFrameBuffer(BOOL enable) {
  if(enable) {
    /* Set up image structure ready to show the framebuffer */
    image.width=truewidth;image.height=trueheight;
    image.xoffset=0; image.format=ZPixmap;
    image.bitmap_unit=depth;
    image.bitmap_pad=depth; // Alignment of bit data
    image.depth=depth; // Depth of display to blit to
    image.bytes_per_line=(depth/8)*truewidth;
    image.bits_per_pixel=depth; // Actual depth of image to blit

    switch(depth) {
    case 8:
      if(!(fbuf=new UBYTE[truewidth*trueheight]))
	return TRUE;
      break;
    case 16:
      if(!(fbuf=new UWORD[truewidth*trueheight]))
	return TRUE;
      break;
    case 24:
      if(!(fbuf=new ULONG[truewidth*trueheight]))
	return TRUE;
      image.bitmap_pad=32;
      image.bits_per_pixel=32;
      image.bytes_per_line=4*truewidth;
      break;
    case 32:
      if(!(fbuf=new ULONG[truewidth*trueheight]))
	return TRUE;
      break;
    default:
      DIE(depth << " bit displays are not currently supported");
      /* Unsupported display depth */
      return FALSE;
    }
    image.data=(char *)fbuf;
    if(XInitImage(&image))
      return TRUE;
  } else {
    delete fbuf;
    fbuf=NULL;
    lastmap.setSize(0,0); // Clear map.
  }
  return FALSE;
}

void Xlib::Draw(TerminalDriver &tty, BOOL refresh) {
  POFUNC;

  if(!fbuf)
    return;

  switch(depth) {
  case 8:
    RenderCharMap8(tty.getCharMap(), lastmap, font, colourmap,
		    (UBYTE *)fbuf, truewidth);
    break;
  case 16:
    RenderCharMap16(tty.getCharMap(), lastmap, font, colourmap,
		    (UWORD *)fbuf, truewidth);
    break;
  case 24:
  case 32:
    RenderCharMap32(tty.getCharMap(), lastmap, font, colourmap,
		    (ULONG *)fbuf, truewidth);
    break;
  default:
    DIE("Attempt to draw to a " << depth << " bit display failed.");
  }
  if(refresh) {
    D("Refreshing whole window");
    XPutImage(dpy, win, gc, &image, 0, 0, 0, 0,
    	      truewidth, trueheight);
  } else {
    for(int Y=0;Y<(tty.getHeight());Y++) {
      if(tty.getCharMap().CompareLines(lastmap, Y)) {
	D("Refreshing line " << Y);
	XPutImage(dpy, win, gc, &image, 0, Y*font.getHeight(), 0, Y*font.getHeight(),
		  truewidth, font.getHeight());
      }
    }
  }
  lastmap=tty.getCharMap();
}

int xerrorhandler(Display *d, XErrorEvent *err) {
  char s[128];
  XGetErrorText(d, err->error_code, s, 128);
  cerr << "X Error: " << s << " (ID=" << err->resourceid << ")\n";
  return 0;
}

Xlib *newXlib(const String &arg) {
  PGFUNC;
  Xlib *d;

  int idx, maxw=0, maxh=0, reqdeep=0;
  String t=arg; // String::before & String::after don't like const input
  if((idx=t.index(":"))>=0) {
    D(t<<" has a : at "<<idx);
    maxw=atoi((String)t.before(idx));
    t=t.after(idx);
    if((idx=t.index(":"))>=0) {
      D(t<<" has a : at "<<idx);
      maxh=atoi((String)t.before(idx));
      t=t.after(idx);
      reqdeep=atoi((String)t);
    } else {
      maxh=atoi((String)t);
    }
  }
  
  D("Parsed settings: W="<<maxw<<", H="<<maxh<<", D="<<reqdeep);

  if(!(maxw&&maxh))
     return NULL;
     
  if(!(d=new Xlib))
    return NULL;
  d->clear();

  d->maxwidth=maxw;
  d->maxheight=maxh;

  if(!(d->dpy=XOpenDisplay(NULL)))
    DIE("Can't open X display " << getenv("DISPLAY"));

  XSetErrorHandler(xerrorhandler);
  d->scr=DefaultScreen(d->dpy);
  if(reqdeep) {
    d->depth=reqdeep;
  } else {
    d->depth=DefaultDepth(d->dpy, d->scr);
  }
  d->cmap=DefaultColormap(d->dpy, d->scr);

  String cnames[16]={
    //    "#000000", "#ff0000", "#00ff00", "#ffff00", "#0000ff", "#ff00ff", "#00ffff", "#ffffff",
    //    "#555555", "#aa5555", "#55aa55", "#aaaa55", "#5555aa", "#aa55aa", "#55aaaa", "#aaaaaa"
    "#000000", "#aa0000", "#00aa00", "#aaaa00", "#0000aa", "#aa00aa", "#00aaaa", "#aaaaaa",
    "#555555", "#ff5555", "#55ff55", "#ffff55", "#5555ff", "#ff55ff", "#55ffff", "#ffffff"
  };

  for(int n=0; n<16; n++) {
    XColor c;
    if(!XAllocNamedColor(d->dpy, d->cmap, cnames[n], &c, &c))
      DIE("Colour map exhausted.");

    d->colourmap[n]=c.pixel;
    D("Viewdata colour " << n << " is X colour " << c.pixel);
  }

  return d;
}

void Xlib::OpenWindow(int width, int height, const String &title) {
  XSizeHints hints;
  hints.width  = hints.min_width  = hints.max_width  = width;
  hints.height = hints.min_height = hints.max_height = height;
  hints.flags  = PSize | PMinSize | PMaxSize;

  XSetWindowAttributes attr;
  attr.background_pixel=BlackPixel(dpy, scr);
  attr.backing_store=WhenMapped;

  D("Opening X window, " << width << "x" << height);
  win=XCreateWindow(dpy, RootWindow(dpy, scr), 0, 0,
		    width, height, 0, CopyFromParent,
		    InputOutput, CopyFromParent,
		    CWBackPixel | CWBackingStore, &attr);

  gc=XCreateGC(dpy, win, 0, 0);
  XSetStandardProperties(dpy, win, title, PROGNAME,
			 0/*iconpixmap*/, NULL, 0, &hints);

  XSelectInput(dpy, win, StructureNotifyMask | ExposureMask | KeyPressMask | KeyReleaseMask );
  XMapRaised(dpy, win);
  XMapWindow(dpy, win);
  XSync(dpy, False);
}

/* predicate procedure used by XCheckIfEvent(), match ALL events !! */
static int all_event_proc(void) {
  return 1;
}

void Xlib::HandleXEvents(ConnectionDriver &sd, TerminalDriver &terminal) {
  XEvent event;
  KeySym keysym;
  char strbuf[16];
  int len;

  while(XCheckIfEvent(dpy, &event, (int(*)(Display *, XEvent *, char *))all_event_proc, NULL)) {
    switch(event.type) {
    case Expose:
      D("Expose event");
      Draw(terminal, TRUE);
      break;

    case KeyPress:
      XLookupString((XKeyEvent *)&event, strbuf, 16, &keysym, NULL);
      D("Keypress keysym " << keysym);
      len=terminal.read(strbuf, 16, keysym);
      if(len>0)
	sd.write(strbuf, len);
      break;

    case KeyRelease:
      XLookupString((XKeyEvent *)&event, strbuf, 16, &keysym, NULL);
      D("Keyrelease keysym " << keysym);
      len=terminal.keyupread(strbuf, 16, keysym);
      if(len>0)
	sd.write(strbuf, len);
      break;

    case UnmapNotify:
      EnableFrameBuffer(FALSE);
      D("Window unmapped");
      break;
    case MapNotify:
      EnableFrameBuffer(TRUE);
      D("Window mapped");
      break;
    case ConfigureNotify:
      D("Configure notification\n");
      break;
    default:
      D("Mystery X event "<< event.type);
    }
  }
}

void Xlib::Main(ConnectionDriver &sd, TerminalDriver &tty) {
  POFUNC;
  struct timeval selwt, when;

  AddFontsFromPath(Settings.fontdir);

  if(font.LoadBin(FindBestFont(maxwidth/tty.getWidth(),
			       maxheight/tty.getHeight())))
    DIE("I don't have a font that fits that size display.");
  if(font.MakeViewdataGlyphs())
    DIE("Couldn't create Viewdata graphics characters");

  truewidth=font.getWidth()*tty.getWidth();
  trueheight=font.getHeight()*tty.getHeight();

  OpenWindow(truewidth, trueheight, PROGNAME " " VERSION /*" - "+title*/);

  char *netbuf=new char[Settings.netbuf];
  int netcount;
  
  fd_set rfd_set;

  int xfd=ConnectionNumber(dpy);

  TickAction(&when, tty);
  for(;;) {
    FD_ZERO(&rfd_set);
    FD_SET(sd.getUnderlyingFD(), &rfd_set);
    FD_SET(xfd, &rfd_set);

    if(QLength(dpy))
      HandleXEvents(sd, tty);

    gettimeofday(&selwt, NULL);
    selwt.tv_sec=when.tv_sec-selwt.tv_sec;
    selwt.tv_usec=when.tv_usec-selwt.tv_usec;
    if(selwt.tv_usec<0) {
      selwt.tv_sec-=selwt.tv_usec/1000000;
      selwt.tv_usec+=(selwt.tv_usec/1000000)*1000000;
    }

    switch(select((sd.getUnderlyingFD()>?xfd)+1,
		  &rfd_set, NULL, NULL, &selwt)) {
    case -1:
      /* Select failed */
      DIE("select() failed because " << strerror(errno));
      break;
    case 0:
      /* Timeout */
      D("select() timed out - ticking");
      TickAction(&when, tty);
      break;
    default:
      /* Activity on fds */
      D("select() got activity on fd");
      if(FD_ISSET(xfd, &rfd_set)) {
	D("Detected X event");
	HandleXEvents(sd, tty);
      }
      if(FD_ISSET(sd.getUnderlyingFD(), &rfd_set)) {
	D("Detected network event");
	netcount=sd.read(netbuf, Settings.netbuf);
	if(netcount<0) {
	  cerr << "Read error from server: " << strerror(errno) << "\n";
	  break;
	} else if(netcount==0) {
	  D("EOF on stream");
	  return;
	}
	tty.write(netbuf, netcount);
	Draw(tty, FALSE);
      }
    }
  }
}

#endif /* USE_XLIB */
